@file:Suppress("MemberVisibilityCanBePrivate", "unused")

package com.ilharper.droidup

import android.annotation.SuppressLint
import android.app.Activity
import android.app.AlertDialog
import android.app.Notification
import android.app.NotificationChannel
import android.app.NotificationManager
import android.app.Service
import android.content.ComponentName
import android.content.Context
import android.content.Intent
import android.content.ServiceConnection
import android.net.Uri
import android.os.Binder
import android.os.Build
import android.os.IBinder
import android.util.Base64
import android.widget.Toast
import androidx.core.app.NotificationCompat
import androidx.core.app.NotificationManagerCompat
import androidx.core.content.FileProvider
import org.json.JSONObject
import java.io.File
import java.io.FileInputStream
import java.io.FileOutputStream
import java.net.HttpURLConnection
import java.net.URL
import java.security.DigestInputStream
import java.security.MessageDigest
import kotlin.math.roundToInt

class DroidUpFileProvider : FileProvider()

/**
 * The DroidUp context.
 *
 * Acts as a bridge to pass data in the entire DroidUp process.
 *
 * @property info The update info.
 * @property activity The activity context.
 * @property file The downloaded apk file.
 * @property manual Whether to show toast when processing.
 */
data class DroidUpContext(
    // Core fields
    var info: DroidUpInfo? = null,
    val activity: Activity,
    var file: File? = null,

    // Options
    val manual: Boolean
)

/**
 * The update info.
 *
 * @property version The versionCode value of the app.
 * @property url The download url.
 * @property hash The hash (SHA-256 by default) of the apk file.
 * @property size The size (in bytes) of the apk file.
 * @property releaseNote The release note which will be displayed in update prompt dialog.
 * @property extra Extra field, where you can assign any type of object.
 */
data class DroidUpInfo(
    val version: Long,
    val url: String?,
    val hash: String?,
    val size: Long?,
    val releaseNote: String?,
    val extra: Any?
)

/**
 * The main DroidUp class.
 */
class DroidUp
private constructor() {

    // region Handlers

    private lateinit var checker: (DroidUpContext) -> DroidUpInfo
    private lateinit var comparator: (DroidUpContext) -> Boolean
    private lateinit var prompter: (DroidUpContext, () -> Unit) -> Unit
    private lateinit var downloader: (DroidUpContext, (File) -> Unit) -> Unit
    private lateinit var verifier: (DroidUpContext) -> Boolean
    private lateinit var installer: (DroidUpContext) -> Unit

    // endregion

    // region Core Methods

    /**
     * Check update.
     *
     * @param activity The activity context. Leave blank to automatically use the current activity.
     * @param manual Whether to show toast when processing.
     */
    fun check(
        activity: Activity? = getActivity(),
        manual: Boolean = false
    ): DroidUp = apply {
        if (activity == null)
            throw IllegalArgumentException("Activity is null")

        // Initialize DroidUp context
        val dctx = DroidUpContext(
            activity = activity,
            manual = manual
        )

        // Runs on new thread
        Thread {
            try {
                // If manual checking, toast start
                if (manual) activity.runOnUiThread {
                    Toast.makeText(activity, R.string.droidup_prompt_check, Toast.LENGTH_SHORT)
                        .show()
                }

                // Do checking
                val info = checker(dctx)
                dctx.info = info

                // Compare version,   then prompt user
                if (comparator(dctx)) prompter(dctx) {
                    // Confirmed, download update
                    downloader(dctx) {
                        dctx.file = it

                        // Verify downloaded file, finally install
                        if (verifier(dctx)) installer(dctx)
                        // Verification failed
                        else activity.runOnUiThread {
                            Toast.makeText(activity, R.string.droidup_verify_failed, Toast.LENGTH_SHORT)
                                .show()
                        }
                    }
                }
                // Already the latest version
                else if (manual) activity.runOnUiThread {
                    Toast.makeText(activity, R.string.droidup_prompt_already_latest, Toast.LENGTH_SHORT)
                        .show()
                }
            } catch (e: Exception) {
                e.printStackTrace()
            }
        }.start()
    }

    // endregion

    // region Builder

    /**
     * The [DroidUp] builder.
     */
    class Builder
    internal constructor() {

        // region Handlers

        var checker: ((DroidUpContext) -> DroidUpInfo)? = null
        var comparator: ((DroidUpContext) -> Boolean)? = null
        var prompter: ((DroidUpContext, () -> Unit) -> Unit)? = null
        var downloader: ((DroidUpContext, (File) -> Unit) -> Unit)? = null
        var verifier: ((DroidUpContext) -> Boolean)? = null
        var installer: ((DroidUpContext) -> Unit)? = null

        // endregion

        /**
         * Build [DroidUp] instance.
         */
        fun build(): DroidUp {
            if (checker == null) throw NullPointerException("Checker is null")
            if (comparator == null) useSimpleComparator()
            if (prompter == null) useSimplePrompter()
            if (downloader == null) useSimpleDownloader()
            if (verifier == null) useSimpleVerifier()
            if (installer == null) useSimpleInstaller()

            return DroidUp().apply {
                checker = this@Builder.checker!!
                comparator = this@Builder.comparator!!
                prompter = this@Builder.prompter!!
                downloader = this@Builder.downloader!!
                verifier = this@Builder.verifier!!
                installer = this@Builder.installer!!
            }
        }

        // region Default Handlers

        /**
         * Use simple checker.
         *
         * This checker fetches `latest.json` from the given [url] and then parse [DroidUpInfo].
         */
        fun useSimpleChecker(url: String) = apply {
            checker = {
                // Parse URL
                val ourl =
                    if (!url.endsWith(".json") && !url.endsWith('/'))
                        "$url/"
                    else url
                var rurl = URL(ourl)
                rurl.apply {
                    if (!path.endsWith(".json"))
                        rurl = URL(rurl, "./latest.json")
                }

                var connection: HttpURLConnection? = null

                try {
                    connection = rurl.openConnection() as HttpURLConnection
                    connection.setRequestProperty("Accept", "application/json")
                    connection.requestMethod = "GET"
                    connection.connect()

                    if (connection.responseCode != HttpURLConnection.HTTP_OK) throw Exception("HTTP error")

                    val json = JSONObject(connection.inputStream.bufferedReader().readText())
                    DroidUpInfo(
                        json.getLong("version"),
                        json.optString("url").let {
                            // Pass absolute url directly,   otherwise convert relative url to absolute
                            if (it.contains("://")) it else URL(rurl, it).toString()
                        },
                        json.optString("hash"),
                        json.optLong("size"),
                        json.optString("releaseNote"),
                        null
                    )
                } catch (e: Exception) {
                    it.activity.runOnUiThread {
                        Toast.makeText(it.activity, R.string.droidup_prompt_check_failed, Toast.LENGTH_SHORT).show()
                    }
                    throw e
                } finally {
                    connection?.disconnect()
                }
            }
        }

        /**
         * Use default comparator. Called by default.
         *
         * This comparator emits update only when [DroidUpInfo.version] > current version.
         */
        fun useSimpleComparator() = apply {
            comparator =
                {
                    it.info!!.version > it.activity.packageManager.getPackageInfo(
                        it.activity.packageName,
                        0
                    ).versionCode
                }
        }

        /**
         * Use default prompter. Called by default.
         *
         * This prompter uses [AlertDialog].
         */
        fun useSimplePrompter() = apply {
            prompter = { dctx, it ->
                dctx.activity.apply {
                    runOnUiThread {
                        AlertDialog.Builder(this).apply {
                            setTitle(R.string.droidup_prompt_title)
                            setMessage(dctx.info!!.releaseNote)
                            setPositiveButton(
                                R.string.droidup_prompt_update
                            ) { _, _ -> it() }
                            setNegativeButton(
                                R.string.droidup_prompt_ignore,
                                null
                            )
                            show()
                        }
                    }
                }
            }
        }

        /**
         * Use default downloader. Called by default.
         *
         * This downloader shows a notification with progress when downloading.
         */
        fun useSimpleDownloader() = apply {
            downloader = { dctx, completed ->
                dctx.activity.bindService(
                    Intent(dctx.activity, SimpleDownloaderService::class.java),
                    object : ServiceConnection {
                        override fun onServiceConnected(name: ComponentName?, service: IBinder?) {
                            (service as SimpleDownloaderService.SimpleDownloaderBinder).start(
                                dctx,
                                completed
                            )
                        }

                        override fun onServiceDisconnected(name: ComponentName?) {
                        }
                    },
                    Context.BIND_AUTO_CREATE
                )
            }
        }

        /**
         * Use default verifier. Called by default.
         *
         * This verifier uses base64 encoded SHA-512 hash string.
         */
        fun useSimpleVerifier() = apply {
            verifier = {
                val md = MessageDigest.getInstance("SHA-512")
                val fis = FileInputStream(it.file)
                val dis = DigestInputStream(fis, md)
                while (true) if (dis.read(ByteArray(1024), 0, 1024) == -1) break
                md.digest().contentEquals(Base64.decode(it.info!!.hash, Base64.DEFAULT))
            }
        }

        /**
         * Use default installer. Called by default.
         */
        fun useSimpleInstaller() = apply {
            installer = {
                it.activity.startActivity(
                    if (Build.VERSION.SDK_INT >= Build.VERSION_CODES.Q)
                        Intent(Intent.ACTION_INSTALL_PACKAGE)
                            .apply {
                                data = FileProvider.getUriForFile(
                                    it.activity,
                                    it.activity.packageName + ".droidupprovider",
                                    it.file!!
                                )
                                addFlags(Intent.FLAG_GRANT_READ_URI_PERMISSION)
                                putExtra(Intent.EXTRA_NOT_UNKNOWN_SOURCE, true)
                            }
                    else
                        Intent(Intent.ACTION_VIEW)
                            .apply {
                                if (Build.VERSION.SDK_INT < Build.VERSION_CODES.N) {
                                    setDataAndType(Uri.fromFile(it.file), "application/vnd.android.package-archive")
                                } else {
                                    setDataAndType(
                                        FileProvider.getUriForFile(
                                            it.activity,
                                            it.activity.packageName + ".droidupprovider",
                                            it.file!!
                                        ),
                                        "application/vnd.android.package-archive"
                                    )
                                    addFlags(Intent.FLAG_GRANT_READ_URI_PERMISSION)
                                }
                                addFlags(Intent.FLAG_ACTIVITY_NEW_TASK)
                            }
                )
            }
        }

        // endregion

    }

    // endregion

    companion object {

        /**
         * The default [DroidUp] instance (singleton).
         */
        @SuppressLint("StaticFieldLeak")
        var default: DroidUp? = null

        /**
         * Create a [Builder].
         *
         * @return The [Builder].
         */
        fun create(): Builder = Builder()

        /**
         * Automatically get the current activity.
         *
         * @return The current activity.
         */
        @SuppressLint("PrivateApi")
        internal fun getActivity(): Activity? {
            val activityThreadClass = Class.forName("android.app.ActivityThread")
            val activityThread = activityThreadClass.getMethod("currentActivityThread").invoke(null)
            val activitiesField = activityThreadClass.getDeclaredField("mActivities").apply {
                isAccessible = true
            }
            for (activityRecord in (activitiesField.get(activityThread) as Map<*, *>).values) {
                val activityRecordClass = activityRecord!!.javaClass
                val pausedField = activityRecordClass.getDeclaredField("paused").apply {
                    isAccessible = true
                }
                if (!pausedField.getBoolean(activityRecord)) {
                    val activityField = activityRecordClass.getDeclaredField("activity").apply {
                        isAccessible = true
                    }
                    return activityField.get(activityRecord) as Activity
                }
            }
            return null
        }

    }

}

/**
 * Droid-Up!
 *
 * The fastest way to create a [DroidUp] instance. Just call with a block to customize the update process.
 */
inline fun droidUp(init: DroidUp.Builder.() -> Unit): DroidUp {
    val builder = DroidUp.create()
    builder.init()
    return builder.build()
}

class SimpleDownloaderService : Service() {

    // region Overrides

    override fun onCreate() {
        notificationManager = NotificationManagerCompat.from(this)
    }

    override fun onBind(intent: Intent): IBinder = binder

    // endregion

    // region Binder

    private val binder = SimpleDownloaderBinder()

    inner class SimpleDownloaderBinder : Binder() {
        fun start(
            dctx: DroidUpContext,
            completed: (File) -> Unit
        ) = apply {
            createNotification()
            Thread {
                download(
                    dctx,
                    completed
                )
            }.start()
        }
    }

    // endregion

    // region Downloader

    private fun download(
        dctx: DroidUpContext,
        completed: (File) -> Unit
    ) {
        val info = dctx.info
        if (info == null) {
            notificationManager.cancel(NOTIFY_ID)
            return
        }

        // Parse URL
        val durl = URL(info.url)
        var path = durl.path
        path = info.url!!.split(path)[0] + path
        if (path.isNullOrEmpty()) throw Exception("Downloader url error")
        val filename = path.split('/').last()

        var connection: HttpURLConnection? = null

        try {
            connection = durl.openConnection() as HttpURLConnection
            connection.requestMethod = "GET"
            connection.connect()

            if (connection.responseCode != HttpURLConnection.HTTP_OK) throw Exception("HTTP error")

            val input = connection.inputStream
            val total = info.size ?: connection.contentLength.toLong()

            // Create temp file
            val droidupProviderFolder = File(applicationContext.cacheDir, "droidup_files")
            droidupProviderFolder.mkdirs()
            val file = File(droidupProviderFolder, filename)
            val output = FileOutputStream(file)

            // Download
            val buffer = ByteArray(1024 * 512)
            var downloaded = 0L
            var tDownloaded = 0L
            var ti = System.currentTimeMillis()

            while (true) {
                val l = input.read(buffer)
                if (l == -1) break
                output.write(buffer, 0, l)
                downloaded += l

                val t = System.currentTimeMillis()
                val dt = t - ti
                if (dt < 2000) continue

                val progress = if (total >= downloaded) (downloaded * 100 / total.toDouble()).roundToInt() else -1
                val td = downloaded - tDownloaded

                // Calculate speed
                var speed = td / dt.toDouble() * 1000
                val speedText = let {
                    if (speed < 1000)
                        return@let String.format("%.1f", speed) + "B/s"
                    speed /= 1024
                    if (speed < 1000)
                        return@let String.format("%.1f", speed) + "KB/s"
                    speed /= 1024
                    return@let String.format("%.1f", speed) + "MB/s"
                }

                ti = t
                tDownloaded = downloaded

                notificationBuilder?.apply {
                    if (progress >= 0) {
                        setContentText("$progress - $speedText")
                        setProgress(100, progress, false)
                    } else {
                        setContentText(speedText)
                        setProgress(0, 0, true)
                    }

                    notificationManager.notify(NOTIFY_ID, build())
                }
            }

            input.close()
            output.close()

            notificationManager.cancel(NOTIFY_ID)

            completed(file)
        } catch (e: Exception) {
            e.printStackTrace()
            dctx.activity.runOnUiThread {
                Toast.makeText(applicationContext, R.string.droidup_download_failed, Toast.LENGTH_SHORT).show()
            }
        } finally {
            notificationManager.cancel(NOTIFY_ID)
            connection?.disconnect()
        }
    }

    // endregion

    // region Notification

    companion object {
        private const val CHANNEL_ID = "droidup_notification"
        private const val NOTIFY_ID = 0
    }

    private lateinit var notificationManager: NotificationManagerCompat
    private var notificationBuilder: NotificationCompat.Builder? = null

    private fun createNotification() {
        if (Build.VERSION.SDK_INT >= Build.VERSION_CODES.O)
            NotificationChannel(
                CHANNEL_ID,
                getString(R.string.droidup_notification_name),
                NotificationManager.IMPORTANCE_HIGH
            ).apply {
                description = getString(R.string.droidup_notification_description)
                lockscreenVisibility = Notification.VISIBILITY_PUBLIC
                setShowBadge(true)
                enableVibration(false)
                enableLights(false)

                notificationManager.createNotificationChannel(this)
            }

        notificationBuilder = NotificationCompat.Builder(this, CHANNEL_ID)
            .apply {
                setContentTitle(getString(R.string.droidup_notification_title))
                setContentText(getString(R.string.droidup_notification_starting))
                setSmallIcon(R.drawable.ic_droidup_update)
                setProgress(0, 0, true)
                priority = NotificationCompat.PRIORITY_DEFAULT

                notificationManager.notify(NOTIFY_ID, build())
            }
    }

    // endregion

}
